(** Time-stamp: <modified the 09/12/2008 (at 17:49) by Erwan Jahier> *)



(* maps node_key to a string that won't clash *)
let node_key_tbl = Hashtbl.create 0

(* maps node name (string) to a counter *)
let node_name_tbl = Hashtbl.create 0

(* exported *)
let (node_key: Eff.node_key -> string -> string) =
  fun nk name -> 
    let (long, sargs) = nk in
      assert (name <> "");
      if sargs = [] then 
        (* If there is no static argument, we don't have to invent a new name,
           and therefore there is nothing to do to avoid clashes *)
        name
      else
        try Hashtbl.find node_key_tbl nk
          (* Note that we ignore the "name" in argument in this case *)
        with 
            Not_found -> 
              (* let's build an ident that won't clash *)
              (* all new name should not begins with a "_" ; hence we prefix by "n_" *)
              let name = "n_" ^ name in
                if not (Hashtbl.mem node_name_tbl name) then
                  (
                    (* that name won't clash, but let's tabulate it *)
                    Hashtbl.add node_name_tbl name 2;
                    Hashtbl.add node_key_tbl nk name;
                    name
                  )
                else
                  (* That name have already been given, there is a possible clash! *)
                  let cpt = Hashtbl.find node_name_tbl name in
                  let fresh_name = 
                    Hashtbl.replace node_name_tbl name (cpt+1);
                    name ^ "_" ^ (string_of_int cpt)
                  in
                    Hashtbl.add node_key_tbl nk fresh_name;
                    fresh_name



(********************************************************************************)
(* Dealing with fresh local (to the node) variable idents *)
let local_var_tbl = Hashtbl.create 0

(* exported *)
let (reset_local_var_prefix : string -> unit) =
  fun str -> 
    Hashtbl.remove local_var_tbl str

(* exported *)
let (new_local_var : string -> string) =
  fun prefix ->
    try 
      let cpt = Hashtbl.find local_var_tbl prefix in
        Hashtbl.replace local_var_tbl prefix (cpt+1);
        "_" ^ prefix ^ (string_of_int cpt)
    with
        Not_found -> 
          Hashtbl.add local_var_tbl prefix 2;
          "_" ^ prefix ^ "1"
    
(********************************************************************************)

let (array_type : Eff.type_ -> string -> string) =
  fun t name -> 
    "A_" ^ name
